"""Client library for using OAuth2, with applications without web access."""

__author__ = 'Luis Maia <luis.maia@xfel.eu>'
__version__ = '6.1.0'

from .oauth2_client_backend import Oauth2ClientBackend
